/* 2D Affine Transformations */
/* Defining Numerical Approximation of PI */
#define PI 3.141592653589793238462643383279502884197
/* End Defining Numerical Approximation of PI */


double size(double v1[2]){
  double dx = v1[0];
  double dy = v1[1];
  return sqrt((dx*dx+dy*dy));
}

double size2(double v1[2]){
  double dx = v1[0];
  double dy = v1[1];
  return ((dx*dx+dy*dy));
}



double dist(double v1[2],double v2[2]){
  double vec[2];
  vec[0] = v1[0]-v2[0];
  vec[1] = v1[1]-v2[1];
  return size(vec);
}


/* set Matrix */

void Scaling(double A[2][2],double xfac,double yfac){
  A[0][0]=xfac;A[0][1]=0.0;
  A[1][0]=0.0;A[1][1]=yfac;
}
 
void Shearing(double A[2][2],double xfac,double yfac){
  A[0][0]=1.0;A[0][1]=xfac;
  A[1][0]=yfac;A[1][1]=1.0;
}

void Line_Inversion(double A[2][2],double mxy[2]){
  double msize = size2(mxy);
  if(msize<=0.0){
    printf("Warning !; your Line_Inversion vector has zero size !\n");
    msize=1.0;
  }
  msize = 1.0/msize;
  double mxmy = msize*2.0*mxy[0]*mxy[1];
  
  A[0][0]=msize*(mxy[0]*mxy[0]-mxy[1]*mxy[1]);A[0][1]=mxmy;
  A[1][0]=mxmy;A[1][1]=-A[0][0];
}

void Rotation(double A[2][2],double theta){
  double angle_rad = (theta*PI)/180.0;
  A[0][0] = cos(angle_rad); A[0][1] = -sin(angle_rad);
  A[1][0] = -A[0][1]; A[1][1] = A[0][0];
}

void Matrix_Matrix_Multiplication(double inX[2][2],double A[2][2],double outX[2][2]){
  int i,j,k;
  for(i=0;i<2;i++)
    for(j=0;j<2;j++){
      outX[i][j]=0.0;
      for(k=0;k<2;k++){
	outX[i][j]+=A[i][k]*inX[k][j];
      }
    }
}
 


/* Forward Mapping */

void Translate(double inX[2],double t[2],double outX[2]){
  outX[0] = inX[0]+t[0];
  outX[1] = inX[1]+t[1];
}
void Vector_Matrix_Multiplication(double inX[2],double A[2][2],double outX[2]){
  outX[0] = A[0][0]*inX[0]+A[0][1]*inX[1];
  outX[1] = A[1][0]*inX[0]+A[1][1]*inX[1];
}

double Determinant(double A[2][2]){
  return A[0][0]*A[1][1]-A[0][1]*A[1][0];
}
void Vector_Matrix_Multiplication_Inverse(double inX[2],double A[2][2],double outX[2]){
  double det = Determinant(A);
  if(det==0.0){
    printf("Warning !; your matrix A does not have an inverse matrix !\n");
    det =1.0;
  }
  
  outX[0] = (A[1][1]*inX[0]-A[0][1]*inX[1])/det;
  outX[1] = (-A[1][0]*inX[0]+A[0][0]*inX[1])/det;
}
void AffineTransformation_FW(double inX[2],double A[2][2],double t[2],double outX[2]){
  /*
    outX[0] = A[0][0]*inX[0]+A[0][1]*inX[1]+t[0];
    outX[1] = A[1][0]*inX[0]+A[1][1]*inX[1]+t[1];
    // or 
    */
  double tmp[2];
  Vector_Matrix_Multiplication(inX,A,tmp);
  Translate(tmp,t,outX);
}

//center coordinates
void AffineTransformation_FW(double inX[2],double A[2][2],double t[2],double ci[2],double co[2],double outX[2]){
  double tmp1[2],tmp2[2],tmp3[2];
  double mc[2];mc[0]=-ci[0];mc[1]=-ci[1];
  Translate(inX,mc,tmp1);
  Vector_Matrix_Multiplication(tmp1,A,tmp2);
  Translate(tmp2,co,tmp3);
  Translate(tmp3,t,outX);
  
}

int getImage_Coordinate(int sy,double v){
  int di = ((int)(v));if(fabs(v-di)>0.5)di++;
  if(di<0)di=0;if(di>sy-1)di=sy-1;
  return di;
}



/* End Foward Mapping */



/* Estimating Size: */
void setNewSize(int sx,int sy,double A[2][2],double t[2],double new_size[2]){
  double q[4][2];
  q[0][0] = 0.0;q[0][1] = 0.0;
  q[1][0] = ((double)(sx));q[1][1] = 0.0;
  q[2][0] = 0.0;q[2][1] = ((double)(sy));
  q[3][0] = ((double)(sx));q[3][1] = ((double)(sy));
  double p[4][2];
  AffineTransformation_FW(q[0],A,t,p[0]);
  AffineTransformation_FW(q[1],A,t,p[1]);
  AffineTransformation_FW(q[2],A,t,p[2]);
  AffineTransformation_FW(q[3],A,t,p[3]);
  double minX,maxX,minY,maxY;
  minX = maxX = p[0][0];
  minY = maxY = p[0][1];
  int i;
  for(i=1;i<4;i++){
    if(maxX<p[i][0])maxX=p[i][0];
    if(minX>p[i][0])minX=p[i][0];
    if(maxY<p[i][1])maxY=p[i][1];
    if(minY>p[i][1])minY=p[i][1];
  }
  new_size[0] = fabs(maxX-minX)+fabs(t[0]);
  new_size[1] = fabs(maxY-minY)+fabs(t[1]);
}




/* End Estimating Size: */

/* Backward Mapping */
void AffineTransformation_BW(double inX[2],double A[2][2],double t[2],double outX[2]){
  
  double tmp[2];
  double mt[2];mt[0]=-t[0];mt[1]=-t[1];
  Translate(inX,mt,tmp);
  Vector_Matrix_Multiplication_Inverse(tmp,A,outX);
  
}
//center coordinates
void AffineTransformation_BW(double inX[2],double A[2][2],double t[2],double ci[2],double co[2],double outX[2]){
    
  double tmp1[2],tmp2[2],tmp3[2];
  
  double mt[2];mt[0]=-(t[0]+co[0]);mt[1]=-(t[1]+co[1]);
  
  Translate(inX,mt,tmp1);
  Vector_Matrix_Multiplication_Inverse(tmp1,A,tmp2);
  Translate(tmp2,ci,outX);
  
}

/* End Backward Mapping */

/* Undefining Numerical Approximation of PI */
#undef PI
/* End Undefining Numerical Approximation of PI */
